
package hipi.compiler;


public class SimpleNode implements Node {
  protected Node parent;
  protected Node[] children;
  protected int id;
  protected HiPiCompiler parser;

  public int line;

  public SimpleNode(int i, int l) {
    id = i;
    line = l;
  }

  public SimpleNode(HiPiCompiler p, int i, int l) {
    this(i, l);
    parser = p;
  }

  public void jjtOpen() {
  }

  public void jjtClose() {
  }
  
  public void jjtSetParent(Node n) { parent = n; }
  public Node jjtGetParent() { return parent; }

  public void jjtAddChild(Node n, int i) {
    if (children == null) {
      children = new Node[i + 1];
    } else if (i >= children.length) {
      Node c[] = new Node[i + 1];
      System.arraycopy(children, 0, c, 0, children.length);
      children = c;
    }
    children[i] = n;
  }

  public Node jjtGetChild(int i) {
    return children[i];
  }

  public int jjtGetNumChildren() {
    return (children == null) ? 0 : children.length;
  }

  // added by Paolo
  // removes the child at position i
  public Node removeChild(int i) {
  	if ((children == null) || (i>=children.length))
  		return null;
  	else if (children.length>1) {
  		Node c[] = new Node[children.length-1];
  		System.arraycopy(children, 0, c, 0, i);
  		System.arraycopy(children, i+1, c, i, c.length-i);
  		Node tmp = children[i];
  		children = c;
  		return tmp;
  	}
  	else {
  		Node tmp = children[0];
  		children = null;
  		return tmp;	
  	}
  }
  
  // added by Paolo
  // inserts a child at position i
  // children at indexes i .. length-1 are shifted
  public void insertChild(Node n, int i) {
	if (children == null) {
	  	children = new Node[i + 1];
	} 
	else if (i >= children.length) {
	  	Node c[] = new Node[i + 1];
	  	System.arraycopy(children, 0, c, 0, children.length);
	  	children = c;
	} 
	else { 
		Node c[] = new Node[children.length + 1];
		System.arraycopy(children, 0, c, 0, i);
		System.arraycopy(children, i, c, i+1, children.length-i);
		children = c;
	}
	children[i] = n;
  }

  /** Accept the visitor. **/
  public Object jjtAccept(HiPiCompilerVisitor visitor, Object data) {
    return visitor.visit(this, data);
  }

  /** Accept the visitor. **/
  public Object childrenAccept(HiPiCompilerVisitor visitor, Object data) {
    if (children != null) {
      for (int i = 0; i < children.length; ++i) {
        children[i].jjtAccept(visitor, data);
      }
    }
    return data;
  }

  /* You can override these two methods in subclasses of SimpleNode to
     customize the way the node appears when the tree is dumped.  If
     your output uses more than one line you should override
     toString(String), otherwise overriding toString() is probably all
     you need to do. */

  public String toString() { return HiPiCompilerTreeConstants.jjtNodeName[id]; }
  public String toString(String prefix) { return prefix + toString(); }

  /* Override this method if you want to customize how the node dumps
     out its children. */

  public void dump(String prefix) {
    System.out.println(toString(prefix));
    if (children != null) {
      for (int i = 0; i < children.length; ++i) {
	SimpleNode n = (SimpleNode)children[i];
	if (n != null) {
	  n.dump(prefix + " ");
	}
      }
    }
  }
}

