/*
 * Created on 9-lug-2003
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package hipi.compiler.encodings;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;

import hipi.compiler.*;
import hipi.compiler.checker.TreeChecker;

/**
 * 
 * @author Paolo Milazzo
 */
public class AddForwardersVisitor implements HiPiCompilerVisitor, HiPiCompilerTreeConstants {

	private static final int FWD_RESERVED_ENV_IDX = 4;
	
	private Vector colocations = new Vector();

	public Object visit(SimpleNode node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTHiPiProgram node, Object data) {
		AddForwardersData recvInfo = new AddForwardersData();
		node.childrenAccept(this,recvInfo);
		return null;
	}

	public Object visit(ASTGlobalChannelDeclaration node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTRecursiveDefinition node, Object data) {
		childrenEncode(node, ((AddForwardersData) data).getCopy()); 
		return null;
	}

	public Object visit(ASTSchedule node, Object data) {
		colocations = node.colocations;
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTScheduleMain node, Object data) {
		childrenEncode(node, ((AddForwardersData) data).getCopy());
		return null;
	}

	public Object visit(ASTBlock node, Object data) {
		childrenEncode(node, ((AddForwardersData) data).getCopy());
		return null;
	}

	public Object visit(ASTLocalDeclarationStatement node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTSend node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTRecv node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTRecursiveDefinitionCall node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTSpawnStatement node, Object data) {
		childrenEncode(node, ((AddForwardersData) data).getCopy());
		return null;
	}

	public Object visit(ASTSelectionStatement node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTReturnStatement node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTExpOp node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTExpIdentifier node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTExpInteger node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTExpString node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTNewChannel node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	// TODO rewrite... the encoding is different from the "foundations" paper
	private void childrenEncode(SimpleNode node, AddForwardersData data) {
		SimpleNode child;
		// -1 is an optimization: a send command at the end
		// of a code fragment has no continuation, so it is
		// already asynchronous.
		for (int i=0; i<node.jjtGetNumChildren() ; i++) {
			child = (SimpleNode) node.jjtGetChild(i);
			
			/*
			 * Encoding:
			 * [x.recv(y);P] = (z)(x-oz|z.recv(y);[P])  if x was received
	 		 *                  x.recv(y);[P]           otherwise
			 */
			 if (ASTRecv.class.isInstance(child)) {
			 	ASTRecv astRecv = (ASTRecv) child;
			 	 
				// execute encoding only on global (idx==-1) 
				// and not reserved (>=ENV_START_IDX) channels
				if ((astRecv.envIndex==-1) || (astRecv.envIndex>=TreeChecker.ENV_START_IDX)) {
					// if isPresent and !negative or !isPresent and negative
					// or is global and not colocated
						if ((isPresent(astRecv.envIndex, data.recvdChannels) != data.negative)
							|| ((astRecv.envIndex==-1) 
								&& (!colocations.contains(astRecv.name)))) {
					
						// create node for (z) 
						ASTLocalDeclarationStatement astLD =
							new ASTLocalDeclarationStatement(JJTLOCALDECLARATIONSTATEMENT);
						astLD.jjtSetParent(node); // paranoid
						astLD.envIndex = FWD_RESERVED_ENV_IDX;
						astLD.type = Types.CHANNEL;
						astLD.params = astRecv.params;
						astLD.TC_visited = true; // paranoid
						astLD.name = null; 
						ASTNewChannel astNew = new ASTNewChannel(JJTNEWCHANNEL);
						astNew.jjtSetParent(astLD); // paranoid
						astLD.jjtAddChild(astNew,0);
					
						// create | node and the forwarder
						ASTSpawnStatement astSpawn = new ASTSpawnStatement(JJTSPAWNSTATEMENT);
						astSpawn.jjtSetParent(node); // paranoid
						astSpawn.CG_visited = false;
						astSpawn.isMigration = false;
						astSpawn.nameGiven = false;
						ASTForwarder astFwd = new ASTForwarder(-1);
						astFwd.jjtSetParent(astSpawn);
						ASTExpIdentifier astId1 = new ASTExpIdentifier(JJTEXPIDENTIFIER);
						astId1.jjtSetParent(astFwd); // paranoid
						astId1.envIndex = astRecv.envIndex;
						astId1.name = astRecv.name;
						ASTExpIdentifier astId2 = new ASTExpIdentifier(JJTEXPIDENTIFIER);
						astId2.jjtSetParent(astFwd); // paranoid
						astId2.envIndex = FWD_RESERVED_ENV_IDX;
						astFwd.jjtAddChild(astId1,0);
						astFwd.jjtAddChild(astId2,1);
						astSpawn.jjtAddChild(astFwd,0);
						
						// change x.recv(y) into z.recv(y)
						astRecv.envIndex = FWD_RESERVED_ENV_IDX;
						astRecv.name = null; // for global channels
						
						// putting all together
						node.insertChild(astLD, i);
						node.insertChild(astSpawn, i+1);
						
						i+=2;
					}
				}

				// continue adding all received channels to recvdChannels
				for (int j=0; j<astRecv.params.size(); j++) {
					FormalParamDescriptor desc = 
						(FormalParamDescriptor) astRecv.params.get(j); 
					if ((desc.type == Types.CHANNEL) && (!data.negative)) {
						data.recvdChannels.add(new Integer(desc.envIndex)); 
					}
				}

			 }
			 else child.jjtAccept(this, data);
		}

	}

	// TODO terrible performances
	private boolean isPresent(int elem, HashSet data) {
		Iterator dataIterator = ((HashSet) data).iterator();
		while (dataIterator.hasNext()) {
			if (((Integer)dataIterator.next()).intValue() == elem)
				return true;
		}
		return false;
	}

	public Object visit(ASTTypeDeclaration node, Object data) {
		// TODO Auto-generated method stub
		return null;
	}

	public Object visit(ASTImportDeclaration node, Object data) {
		// TODO Auto-generated method stub
		return null;
	}
}
