package hipi.compiler.encodings;

import hipi.compiler.*;

import java.util.HashSet;
import java.util.Vector;

/*
 * Created on 9-lug-2003
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */

/**
 * 
 * @author Paolo Milazzo
 */
public class SyncToAsyncVisitor implements HiPiCompilerVisitor, HiPiCompilerTreeConstants {

	private static final int RESERVED_ENV_IDX = 3;
	
	// this is used to prevent the visit of nodes added by
	// this encoding.
	// TODO better solution
	private HashSet addedNodes = new HashSet();

	// these nodes are used to send an acknoledgment after the recv.
	// They are unique and used a lot of times. This solution
	// avoids a lot of threads in the output file (optimization)
	private ASTSend astSend = new ASTSend(JJTSEND);
	private ASTSpawnStatement astSpawn = new ASTSpawnStatement(JJTSPAWNSTATEMENT);

	public Object visit(SimpleNode node, Object data) {
		if (!addedNodes.contains(node))
			node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTHiPiProgram node, Object data) {
		// initialization of the nodes used for the acknoledgement
		addedNodes.add(astSpawn);
		addedNodes.add(astSend);
		astSpawn.jjtSetParent(null); // this could be unsafe... in testing
		astSpawn.jjtAddChild(astSend,0);
		astSpawn.nameGiven = false;
		astSpawn.CG_visited = false;
		astSend.envIndex = RESERVED_ENV_IDX;
		astSend.jjtSetParent(astSpawn); 
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTGlobalChannelDeclaration node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTRecursiveDefinition node, Object data) {
		childrenEncode(node);
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTSchedule node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTScheduleMain node, Object data) {
		childrenEncode(node);
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTBlock node, Object data) {
		childrenEncode(node);
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTLocalDeclarationStatement node, Object data) {
		if (!addedNodes.contains(node))
			node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTSend node, Object data) {
		if (!addedNodes.contains(node))
			node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTRecv node, Object data) {
		if (!addedNodes.contains(node))
			node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTRecursiveDefinitionCall node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTSpawnStatement node, Object data) {
		if (!addedNodes.contains(node)) {
			childrenEncode(node); 
			node.childrenAccept(this,data);
		}
		return null;
	}

	public Object visit(ASTSelectionStatement node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTReturnStatement node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTExpOp node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTExpIdentifier node, Object data) {
		if (!addedNodes.contains(node))
			node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTExpInteger node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTExpString node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTNewChannel node, Object data) {
		if (!addedNodes.contains(node))
			node.childrenAccept(this,data);
		return null;
	}

	private void childrenEncode(SimpleNode node) {
		SimpleNode child;
		for (int i=0; i<node.jjtGetNumChildren() ; i++) {
			child = (SimpleNode) node.jjtGetChild(i);
			
		    /*
		     * Encoding:
		     * [x.send(y);P] = (z)(x.send(z,y)|z.recv();[P])
		     */
			if (ASTSend.class.isInstance(child)) {
				
				// create (z) node
				ASTLocalDeclarationStatement astLD = 
					new ASTLocalDeclarationStatement(JJTLOCALDECLARATIONSTATEMENT);
				addedNodes.add(astLD);
				astLD.jjtSetParent(node); // paranoid
				astLD.envIndex = RESERVED_ENV_IDX;
				astLD.type = Types.CHANNEL;
				astLD.params = new Vector();
				astLD.TC_visited = true; // paranoid
				// name=null is useful to make AddForwarderVisitor working properly
				astLD.name = null; 
				ASTNewChannel astNew = new ASTNewChannel(JJTNEWCHANNEL);
				addedNodes.add(astNew);
				astNew.jjtSetParent(astLD); // paranoid
				astLD.jjtAddChild(astNew,0);
				
				// add z to send
				ASTExpIdentifier astId = new ASTExpIdentifier(JJTEXPIDENTIFIER);
				addedNodes.add(astId);
				astId.envIndex = RESERVED_ENV_IDX;
				child.insertChild(astId, 0);
				
				// create | node
				ASTSpawnStatement astSpawn = new ASTSpawnStatement(JJTSPAWNSTATEMENT);
				addedNodes.add(astSpawn);
				astSpawn.jjtSetParent(node); // paranoid
				child.jjtSetParent(astSpawn);
				astSpawn.jjtAddChild(child,0);

				// create recv
				ASTRecv astRecv = new ASTRecv(JJTRECV);
				addedNodes.add(astRecv);
				astRecv.envIndex = RESERVED_ENV_IDX;
				astRecv.params = new Vector(); // no parameters
				
				// putting all together
				node.jjtAddChild(astLD,i); // replace old ASTSend child
				node.insertChild(astSpawn, i+1);
				node.insertChild(astRecv, i+2);
				
				i+=2;
			}
			
			/*
			 * Encoding:
			 * [x.recv(y);P] = x.recv(z,y)(z.send()|[P])
			 */
			else if (ASTRecv.class.isInstance(child)) {
				
				// add z to recv
				if (((ASTRecv) child).params == null)
					((ASTRecv) child).params = new Vector();
				((ASTRecv) child).params.add(0,new FormalParamDescriptor(RESERVED_ENV_IDX));
								
				// putting all together (astSpawn is global)
				node.insertChild(astSpawn, i+1);
				
				i++;

			}
		}
	}

	public Object visit(ASTTypeDeclaration node, Object data) {
		// TODO Auto-generated method stub
		return null;
	}

	public Object visit(ASTImportDeclaration node, Object data) {
		// TODO Auto-generated method stub
		return null;
	}
}
