/*
 * Created on 24-lug-2003
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package hipi.simulator;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Vector;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * 
 * @author Paolo Milazzo
 */
public class SimThread {
	
	// environment of the thread
	// contains Strings, Integers and URIs
	private Vector env = new Vector();
	
	// reference to the root node of the schedule (DOM document)
	private Document schedule;
	
	// list of nodes representing the code fragment
	// contains DOM Elements
	private ArrayList codeFrag = new ArrayList();
	
	// referece to the virtual machine where this thread is executed
	private VirtualMachine vm;
	
	// thread name
	private String name;
	
	// true if terminated
	private boolean terminated;
	
	// true if blocked on a channel
	private boolean blocked;

	// default constructor
	public SimThread(Document sched, Node thread, VirtualMachine machine, Vector fatherEnv) {
		// initialize local variables
		this.terminated = false;
		this.vm = machine;
		this.schedule = sched;
		// initialize thread name
		String schedName = schedule.getFirstChild().getAttributes().getNamedItem("name").getNodeValue();
		String threadName = thread.getAttributes().getNamedItem("name").getNodeValue();
		this.name= machine.getName() + "/" + schedName + "/" + threadName;
		// create codeFrag
		codeFrag = filter(thread.getChildNodes());
		// initialize env
		env = new Vector();
		while (((Element)codeFrag.get(0)).getNodeName().equals("env") 
				&& (fatherEnv != null)) {
			String attr = ((Element)codeFrag.remove(0)).getAttribute("idx");
			if (attr.equals("all")) {
				env=(Vector)fatherEnv.clone();
			}
			else {
				int idx=Integer.parseInt(attr); 
				insertEnv(idx,fatherEnv.get(idx));
			}
		}
	}
	
	// sets the thread terminated or not
	public void setTerminated(boolean term) {
		terminated = term;
	}
	
	// returns true if terminated
	public boolean isTerminated() {
		return terminated;
	}

	// sets the thread terminated or not
	public void setBlocked(boolean block) {
		blocked = block;
	}
	
	// returns true if blocked on a channel
	public boolean isBlocked() {
		return blocked;
	}
	
	// returns thread name
	public String getName() {
		return name;
	}
	
	// inserts one element in the env Vector
	public void insertEnv(int index, Object elem) {
		if (index>=env.size()) { 
			env.setSize(index+1);
			env.add(index,elem);
		}
		else
			env.set(index, elem);
	}

	// executes one step
	// this is the real XIL interpreter
	// TODO use exceptions insead of error messages
	public void step(boolean verbose) {
		
		try {

		
		// if terminated
		if (codeFrag.size()==0) {
			terminated = true;
			print_out(verbose,name + " terminated");
		}
		else {
			// remove next instruction
			Element node = (Element) codeFrag.remove(0);
			String instr = node.getNodeName();
			
			// <newch idx="n"/>
			if (instr.equals("newch")) {
				int idx = Integer.parseInt(node.getAttribute("idx"));
				URI uri = vm.createLocalChannel(name);
				insertEnv(idx,uri);
				print_out(verbose,name + " - newch - " + uri);
			}
			
			// <spawn thread="threadname" dest="ch"/>
			else if (instr.equals("spawn")) {
				String threadName = node.getAttribute("thread");
				NodeList threads = schedule.getElementsByTagName("thread");
				VirtualMachine destVM = vm;
				if (!node.getAttribute("dest").equals("")) {
					try {
						destVM = vm.resolve((URI)load(node.getAttribute("dest")));
					}
					catch (ClassCastException cce) {
						System.out.println("DESTINATION CHANNEL (FOR MIGRATION) NOT VALID");
						System.out.println("IN " + name);
					}
				}
				for (int i=0; i<threads.getLength(); i++) {
					String attribute = ((Element)threads.item(i)).getAttribute("name"); 
					if (attribute.equals(threadName)) {
						SimThread newThread = 
							new SimThread(schedule, threads.item(i), destVM, env);
						destVM.addDepBag(newThread);
					}
				}
				print_out(verbose,name + " - spawn - " + threadName);
			}
			
			// <send> expressions </send>
			else if (instr.equals("send")) {
				ArrayList childNodes = filter(node.getChildNodes());
				URI uri=null;
				if (childNodes.size()==0) {
					System.out.println("NO SUBJECT SPECIFIED FOR <send>");
					System.out.println("IN " + name);
				}
				else {
					Element subj = (Element) childNodes.get(0); 
					if (!subj.getNodeName().equals("load")) {
						System.out.println("SUBJECT INCORRECT FOR <send>");
						System.out.println("IN " + name);
					}
					else {
						String src = subj.getAttribute("src");
						try {
							uri = (URI)load(src);
							if (uri==null) {
								System.out.println("SUBJECT OF <send> NOT IN ENVIRONMENT");
								System.out.println("IN " + name);
							}
						}
						catch (ClassCastException cce) {
								System.out.println("SUBJECT OF <send> IS NOT A CHANNEL");
								System.out.println("IN " + name);
						}
						// if it is an output of a string to console
						if ((uri.equals(Simulator.CONSOLE_STRING_URI)) 
							|| (uri.equals(Simulator.CONSOLE_INT_URI)
							|| (uri.equals(Simulator.CONSOLE_CHANNEL_URI)))) {
							if (childNodes.size()==3) {
								// message is the third child (second one is the ack channel)
								Element child = (Element)childNodes.get(2);
								String childName = child.getNodeName();
								Object msg = null;
								if (childName.equals("load")) {
									msg=load(child.getAttribute("src"));
								}
								else if ((childName.equals("int")) ||
										(childName.equals("string")) ||
										(childName.equals("newch")) ||
										(childName.equals("op"))) {
									msg=evaluate(child);
								}
								else {
									System.out.println("CONSOLE OUTPUT IS NOT VALID");
									System.out.println("IN " + name);
								}
								System.out.println(name + "> " + msg);
								// query dns for ack channel
								// TODO more controls
								URI ack=(URI)load(((Element)childNodes.get(1)).getAttribute("src"));
								VirtualMachine location = vm.resolve(ack);
								location.makeOutput(ack, this, new Vector());								
							}
							else {
								System.out.println("CONSOLE OUTPUT MESSAGE IS NOT WELLFORMED");
								System.out.println("IN " + name);
							}
						}
						// otherwise
						else {
							// query dns for subject channel
							VirtualMachine location = vm.resolve(uri);
							if (location==null) {
								System.out.println("SUBECT OF <send> DOESN'T EXIST IN THE NET");
								System.out.println("IN " + name);
								System.out.println("URI " + uri);
							} 
							else {
								// creates the object of the communication
								Vector message = new Vector();
								for (int i=1; i<childNodes.size(); i++) {
									Element child = (Element)childNodes.get(i);
									String childName = child.getNodeName();
									if (childName.equals("load")) {
										message.add(load(child.getAttribute("src")));
									}
									else if ((childName.equals("int")) ||
											(childName.equals("string")) ||
											(childName.equals("newch")) ||
											(childName.equals("op"))) {
										message.add(evaluate(child));
									}
									else {
										System.out.println("ERROR IN THE OBJECT OF <send>");
										System.out.println("IN " + name);
									}
								}
								location.makeOutput(uri, this, message);
							}
						}
					}
				}
				print_out(verbose,name + " - send - " + uri);
			}

			// <recv> expressions </recv>
			else if (instr.equals("recv")) {
				ArrayList childNodes = filter(node.getChildNodes());
				URI uri=null;
				if (childNodes.size()==0) {
					System.out.println("NO SUBJECT SPECIFIED FOR <recv>");
					System.out.println("IN " + name);
				}
				else {
					Element subj = (Element) childNodes.get(0); 
					if (!subj.getNodeName().equals("load")) {
						System.out.println("SUBJECT INCORRECT FOR <recv>");
						System.out.println("IN " + name);
					}
					else {
						String src = subj.getAttribute("src");
						try {
							uri = (URI)load(src);
							if (uri==null) {
								System.out.println("SUBJECT OF <recv> NOT IN ENVIRONMENT");
								System.out.println("IN " + name);
							}
						}
						catch (ClassCastException cce) {
								System.out.println("SUBJECT OF <recv> IS NOT A CHANNEL");
								System.out.println("IN " + name);
						}
						// if it is an output of a string to console
						if ((uri.equals(Simulator.CONSOLE_STRING_URI)) 
							|| (uri.equals(Simulator.CONSOLE_INT_URI)
							|| (uri.equals(Simulator.CONSOLE_CHANNEL_URI)))) {
							if (childNodes.size()==3) {
								// message is the third child (second one is the ack channel)
								Element child = (Element)childNodes.get(2);
								String childName = child.getNodeName();
								Object msg = null;
								if (childName.equals("store")) {
									if (uri.equals(Simulator.CONSOLE_STRING_URI)) {
										System.out.print(name + "< ");
										msg = Simulator.consoleInput.readLine();
									}
									else if (uri.equals(Simulator.CONSOLE_INT_URI)) {
										System.out.print(name + "< ");
										msg = new Integer(Simulator.consoleInput.readLine().trim());
									}
									else if (uri.equals(Simulator.CONSOLE_CHANNEL_URI)) {
										System.out.print(name + "< ");
										msg = new URI(Simulator.consoleInput.readLine().trim());
									}
								}
								else {
									System.out.println("CONSOLE INPUT IS NOT VALID");
									System.out.println("IN " + name);
								}
								insertEnv(Integer.parseInt(child.getAttribute("idx")),msg);
								// skip next instruction (the spawn of the process sending ack)
								codeFrag.remove(0);
								/*
								URI ack=vm.createLocalChannel(name);
								insertEnv(Integer.parseInt(((Element)childNodes.get(1)).getAttribute("idx")),ack);
								vm.makeInput(ack, this, new Vector());
								*/																
							}
							else {
								System.out.println("CONSOLE INPUT MESSAGE IS NOT WELLFORMED");
								System.out.println("IN " + name);
							}
						}
						// otherwise
						else {
							// query dns for subject channel
							VirtualMachine location = vm.resolve(uri);
							if (location==null) {
								System.out.println("SUBJECT OF <recv> DOESN'T EXIST IN THE NET");
								System.out.println("IN " + name);
								System.out.println("URI " + uri);
							} 
							else {
								// creates the object of the communication
								Vector message = new Vector();
								for (int i=1; i<childNodes.size(); i++) {
									Element child = (Element)childNodes.get(i);
									String childName = child.getNodeName();
									if (childName.equals("store")) {
										message.add(new Integer(child.getAttribute("idx")));
									}
									// discard is not implemented in the compiler
									else if (childName.equals("discard")) {
										message.add(new Integer(-1));					
									}
									else {
										System.out.println("ERROR IN THE OBJECT OF <recv>");
										System.out.println("IN " + name);
									}
								}
								location.makeInput(uri, this, message);
							}
						}
					}
				}
				print_out(verbose,name + " - recv - " + uri);
			}

			// <fwd> expression expression </fwd>
			else if (instr.equals("fwd")) {
				ArrayList childNodes = filter(node.getChildNodes());
				URI uri=null; URI dest=null;
				if (childNodes.size()!=2) {
					// paranoid: this is checked by xml schema validation
					System.out.println("LINEAR FORWARDER NOT WELLFORMED");
					System.out.println("IN " + name);
				}
				else {
					// handle the source of the forwarder
					Element first = (Element) childNodes.get(0); 
					if (!first.getNodeName().equals("load")) {
						System.out.println("FIRST ELEMENT OF <fwd> INCORRECT");
						System.out.println("IN " + name);
					}
					else {
						String src = first.getAttribute("src");
						try {
							uri = (URI)load(src);
							if (uri==null) {
								System.out.println("FIRST ELEMENT OF <fwd> NOT IN ENVIRONMENT");
								System.out.println("IN " + name);
							}
						}
						catch (ClassCastException cce) {
								System.out.println("FIRST ELEMENT OF <fwd> IS NOT A CHANNEL");
								System.out.println("IN " + name);
						}
						// query dns for first channel
						VirtualMachine location = vm.resolve(uri);
						if (location==null) {
							System.out.println("FIRST ELEMENT OF <fwd> DOESN'T EXIST IN THE NET");
							System.out.println("IN " + name);
							System.out.println("URI " + uri);
						} 
						else {
							// handle the destination of the forwarder
							Element second = (Element) childNodes.get(1); 
							if (!second.getNodeName().equals("load")) {
								System.out.println("SECOND ELEMENT OF <fwd> INCORRECT");
								System.out.println("IN " + name);
							}
							else {
								String sndSrc = second.getAttribute("src");
								try {
									dest = (URI)load(sndSrc);
									if (dest==null) {
										System.out.println("SECOND ELEMENT OF <fwd> NOT IN ENVIRONMENT");
										System.out.println("IN " + name);
									}
								}
								catch (ClassCastException cce) {
										System.out.println("SECOND ELEMENT OF <fwd> IS NOT A CHANNEL");
										System.out.println("IN " + name);
								}
								location.makeFwd(uri, this, dest);
							}
						}
					}
				}
				print_out(verbose,name + " - fwd - " + uri + " -o " + dest);
			}
			
			// <store idx="n"> expression </store>
			else if (instr.equals("store")) {
				int idx = Integer.parseInt(node.getAttribute("idx"));
				ArrayList childNodes = filter(node.getChildNodes());
				insertEnv(idx,evaluate((Element)childNodes.get(0)));
				print_out(verbose,name + " - store - " + idx);
			}
			
			// <if> expression <then> P </then> <else> P' <else/> <if/>
			else if (instr.equals("if")) {
				ArrayList childNodes = filter(node.getChildNodes());
				ArrayList cont = null;
				if (((Integer)evaluate((Element)childNodes.get(0))).intValue()!=0)
					cont = filter(((Element)childNodes.get(1)).getChildNodes());
				else
					cont = filter(((Element)childNodes.get(2)).getChildNodes());
				// update codeFrag
				// P or P' is added at the begin of codeFrag
				int j=0;
				for (int i=0; i<cont.size(); i++) {
						codeFrag.add(j++, cont.get(i));
				}
				print_out(verbose,name + " - if ");				
			}
			
			// <terminate/>
			else if (instr.equals("terminate")) {
				terminated = true;
			}
			
			// otherwise
			else {
				print_out(true,name + " " + instr + " not implemented");
			}
		}
		
		}
		catch (Exception e) {
			if (URISyntaxException.class.isInstance(e)) {
				System.out.println("invalid URI: restart schedule");
			}
			else if (NumberFormatException.class.isInstance(e)) {
				System.out.println("malformed integer: restart schedule");
			}
			else {
				System.err.println("EXCEPTION IN " + name);
				e.printStackTrace();
			}
		}
	}


	// loads a global channel or an element of the environment
	private Object load(String src) {
		try {
			// if src is a global channel
			// returns its URI ...
			URI x = new URI(src);
			if (x.isAbsolute()) {
			 	return x;
			}
			else {
				// ... else load from environment
				return env.get(Integer.parseInt(src));
			}
		}
		catch (URISyntaxException e) {
			System.out.println("MALFORMED URI");
			System.out.println("IN " + name);
			return null;
		}
	}

	// filters a NodeList (only Elements are left)
	private ArrayList filter(NodeList childNodes) {
		ArrayList retList = new ArrayList();	
		for (int i=0; i<childNodes.getLength(); i++) {
			if (Element.class.isInstance(childNodes.item(i)))
				retList.add(childNodes.item(i));
		}
		return retList;
	}
	
	// evaluates expressions
	private Object evaluate(Element elem) {
		String nodeName = elem.getNodeName();
		if (nodeName.equals("load")) {
			return load(elem.getAttribute("src"));
		}
		if (nodeName.equals("int")) {
			return new Integer(elem.getFirstChild().getNodeValue().trim());					
		}
		else if (nodeName.equals("string")) {
			if (elem.hasChildNodes()) {
				return elem.getFirstChild().getNodeValue().trim();
			}
			else return new String(); // handle the empty string
		}
		else if (nodeName.equals("newch")) {
			return vm.createLocalChannel(name);
		}
		else if (nodeName.equals("op")) {
			String opType = elem.getAttribute("type");
			ArrayList childNodes = filter(elem.getChildNodes());
			if (childNodes.size() == 2) {
			try {
				// binary operators
				if (opType.equals("add")) {
					Integer ch1 = (Integer)evaluate((Element)childNodes.get(0));
					Integer ch2 = (Integer)evaluate((Element)childNodes.get(1));
					return new Integer(ch1.intValue() + ch2.intValue());
				}
				if (opType.equals("sub")) {
					Integer ch1 = (Integer)evaluate((Element)childNodes.get(0));
					Integer ch2 = (Integer)evaluate((Element)childNodes.get(1));
					return new Integer(ch1.intValue() - ch2.intValue());
				}
				if (opType.equals("mul")) {
					Integer ch1 = (Integer)evaluate((Element)childNodes.get(0));
					Integer ch2 = (Integer)evaluate((Element)childNodes.get(1));
					return new Integer(ch1.intValue() * ch2.intValue());
				}
				if (opType.equals("div")) {
					Integer ch1 = (Integer)evaluate((Element)childNodes.get(0));
					Integer ch2 = (Integer)evaluate((Element)childNodes.get(1));
					return new Integer(ch1.intValue() / ch2.intValue());
				}
				if (opType.equals("mod")) {
					Integer ch1 = (Integer)evaluate((Element)childNodes.get(0));
					Integer ch2 = (Integer)evaluate((Element)childNodes.get(1));
					return new Integer(ch1.intValue() % ch2.intValue());
				}
				if (opType.equals("and")) {
					if (((Integer)evaluate((Element)childNodes.get(0))).intValue()==0)
						return new Integer(0);
					else
						return (Integer)evaluate((Element)childNodes.get(1));
				}
				if (opType.equals("or")) {
					if (((Integer)evaluate((Element)childNodes.get(0))).intValue()!=0)
						return new Integer(1);
					else
						return (Integer)evaluate((Element)childNodes.get(1));
				}
				if (opType.equals("eq")) {
					Object ch1 = evaluate((Element)childNodes.get(0));
					Object ch2 = evaluate((Element)childNodes.get(1));
					return new Integer(ch1.equals(ch2)? 1 : 0);
				}
				if (opType.equals("neq")) {
					Object ch1 = evaluate((Element)childNodes.get(0));
					Object ch2 = evaluate((Element)childNodes.get(1));
					return new Integer(ch1.equals(ch2)? 0 : 1);
				}
				if (opType.equals("lt")) {
					Integer ch1 = (Integer)evaluate((Element)childNodes.get(0));
					Integer ch2 = (Integer)evaluate((Element)childNodes.get(1));
					return new Integer(ch1.intValue() < ch2.intValue()? 1 :0);
				}
				if (opType.equals("gt")) {
					Integer ch1 = (Integer)evaluate((Element)childNodes.get(0));
					Integer ch2 = (Integer)evaluate((Element)childNodes.get(1));
					return new Integer(ch1.intValue() > ch2.intValue()? 1 :0);
				}
				if (opType.equals("le")) {
					Integer ch1 = (Integer)evaluate((Element)childNodes.get(0));
					Integer ch2 = (Integer)evaluate((Element)childNodes.get(1));
					return new Integer(ch1.intValue() <= ch2.intValue()? 1 :0);
				}
				if (opType.equals("ge")) {
					Integer ch1 = (Integer)evaluate((Element)childNodes.get(0));
					Integer ch2 = (Integer)evaluate((Element)childNodes.get(1));
					return new Integer(ch1.intValue() >= ch2.intValue()? 1 :0);
				}
				else {
					System.out.println("INCORRECT NUMBER OF CHILD IN EXPRESSION");
					System.out.println("IN " + name);
				}
			}
			catch (ClassCastException cce) {
				System.out.println("ERROR IN EXPRESSION");
				System.out.println("IN " + name);
			}
			}
			else if (childNodes.size() == 1) {
			try {
				if (opType.equals("umin")) {
					Integer ch = (Integer)evaluate((Element)childNodes.get(0));
					return new Integer(ch.intValue() * (-1));
				}
				if (opType.equals("not")) {
					Integer ch = (Integer)evaluate((Element)childNodes.get(0));
					return new Integer(ch.intValue()==0? 1 :0);
				}
				else {
					System.out.println("INCORRECT NUMBER OF CHILD IN EXPRESSION");
					System.out.println("IN " + name);
				}
			}
			catch (ClassCastException cce) {
				System.out.println("ERROR IN EXPRESSION");
				System.out.println("IN " + name);
			}
			}
		}
		return null;
	}

	// TODO USE THIS METHOD
	private void print_err(String msg) {
		System.out.println(msg);
		System.out.println("IN " + name);
	}
	
	// prints output messages
	private void print_out(boolean verbose, String msg) {
		if (verbose)
			System.out.println(msg);
	}
}
